/* unmangle.c (c) 1990 by Zortech - used by permission
   
   You may distribute or use this file without restriction except that
   the above copyright notice should remain in the source file if it is
   electronically or otherwise distributed.
   
   Zortech ltd and Zortech Inc do not warranty that this software will 
   perform according to the descriptions, and will not be responsible for
   any loss or damage consequential or otherwise, which may occur as a
   result of using these functions.
   
   This file contains a number of functions which allow the un-mangling
   of C++ mangled names as used in Zortech C++ v2. The primary function
   is unmangle_str which takes a character string containing the mangled
   name and returns a pointer to the un-mangled name. The accompanying
   file ownclass.c contains a version which strips out the function
   parameter information - useful for debuggers which will normally obtain
   this from the debugging information in the executable file.
   
   Be warned these routines have not been thoroughly tested and may contain
   obscure bugs.
   
   written by Phil Murray and Dave Mansell of Zortech - March 1990
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


typedef struct unmangle_s {
	char	*name;
	char	*oper;
} UNMANGLE;

static UNMANGLE	um_table[] = {
	{"ct","1"},
	{"dt","2"},
	{"nw"," new"},
	{"dl"," delete"},
	{"pl","+"},
	{"mi","-"},
	{"ml","*"},
	{"dv","/"},
	{"md","%"},
	{"er","^"},
	{"ad","&"},
	{"or","|"},
	{"co","~"},
	{"nt","!"},
	{"as","="},
	{"lt","<"},
	{"gt",">"},
	{"apl","+="},
	{"ami","-="},
	{"amu","*="},
	{"adv","/="},
	{"amd","%="},
	{"aer","^="},
	{"aad","&="},
	{"aor","|="},
	{"ls","<<"},
	{"rs",">>"},
	{"ars",">>="},
	{"als","<<="},
	{"eq","=="},
	{"ne","!="},
	{"le","<="},
	{"ge",">="},
	{"aa","&&"},
	{"oo","||"},
	{"pp","++"},
	{"mm","--"},
	{"cl","()"},
	{"vc","[]"},
	{"rf","->"},
	{"cm",","},
	{"rm","->*"},
	{"op","3"}};


typedef struct paramangle_s {
	char	type;
	char	*desc;
} PARAMANGLE;

static PARAMANGLE type_table[] = {
	{'T',"1"},
/* data types */
	{'c',"char"},
	{'s',"short"},
	{'i',"int"},
	{'l',"long"},
	{'f',"float"},
	{'d',"double"},
	{'r',"long double"},
/* pointers and qualifiers */
	{'v',"void"},
	{'b'," __ss *"},
	{'h'," __handle *"},
	{'p'," near *"},
	{'P'," far *"},
	{'H',"huge "},
	{'R'," &"},
	{'C',"const "},
	{'V',"volatile "},
	{'U',"unsigned "},
/* functions types */
	{'B'," near cdecl func)("},
	{'D'," far cdecl func)("},
	{'N'," near func)("},
	{'F'," far func)("},
/* other bits */
	{'e',"..."},
	{'A',"array of "},
	{'Z',") returning "}};


/*
	unmangle_str

	This function accepts a mangled C++ name and returns a simplified
	unmangled name. Returns NULL if no unmangling required.

	Note this returns a malloced buffer which should be freed when
	finished with if conversion takes place.
*/

char	*unmangle_str(char *str)
{
	char	*unmangle_parm(char *);
	char	*result=NULL, *par=NULL;
	char	*name,*s,*s1,*cp,*op;
	unsigned len,i;

	name = malloc(strlen(str)+1);					   /*  copy for safety  */
	if (name == NULL) {
		printf("unmangle_str: Out of heap\n");
		exit(EXIT_FAILURE);
	}
	strcpy(name,str);
	s1 = strstr(name,"__");									  /*  get '__'  */
	if (s1==name) s1 = strstr(s1+2,"__");		 /*  ignore starting value  */
	if (s1) {									/*  found and not at start  */
		*s1 = '\0';								/*  terminate string early  */
		cp = s1+2;								 /*  search for class name  */
		len = 0;
		while ((*cp>='0')&&(*cp<='9')) {	  /*  get length of class name  */
			len *= 10;
			len += (*(cp++)-'0');
		}
		if (len) {									  /*  class name found  */
			par = unmangle_parm(&cp[len]);
			cp[len] = '\0';								  /*  mark for use  */
		}
		else
		{
			par = unmangle_parm(cp);
			*cp = 0;								/*  else no class name  */
		}
		op = NULL;								  /*  show no operator yet  */
		s = name;
		if (*s=='_') {						   /*  check for special names  */
			s++;
			if (s[0]=='_'&&s[1]=='_') {			  /*  need to search table  */
				for (i=0;i<(sizeof(um_table)/sizeof(struct unmangle_s));i++) {
					if (strncmp(um_table[i].name,s+2,strlen(um_table[i].name))==0) {
						op = um_table[i].oper;
						break;
					}
				}
			}
		}
		if (cp) {			  /*  if class pointer else use shortened name  */
			result = malloc(256);					/*  get largish buffer  */
			if (result == NULL) {
				printf("unmangle_str: Out of heap\n");
				exit(EXIT_FAILURE);
			}
			s1 = result;
			if (strcmp(name,"_base")==0) {		   /*  detect base classes  */
				s = cp;								 /*  make name = class  */
			}
/*			else if (name[0]=='_') *s1++ = '_';	    copy '_' for sorting  */
			if (*cp)
				sprintf(s1,"%s::",cp);
			else
				sprintf(s1,"%s",cp);
			while (*s1) s1++;
			if (op) {
				switch (*op) {
				  case '1':								   /*  constructor  */
					strcpy(s1,cp);
					break;
				  case '2':									/*  destructor  */
					sprintf(s1,"~%s",cp);
					break;
				  default:									  /*  operator  */
					sprintf(s1,"operator%s",(*op=='3')?" <user>":op);
					break;
				}
			}
			else strcat(result,s);					/*  add on actual name  */
		}
		else if (op) {
			result = malloc(256);					/*  get largish buffer  */
			if (result == NULL) {
				printf("unmangle_str: Out of heap\n");
				exit(EXIT_FAILURE);
			}
			switch (*op) {
			  case '1':								   /*  constructor  */
			  case '2':									/*  destructor  */
				break;
			  default:									  /*  operator  */
				sprintf(result,"operator%s",(*op=='3')?" <user>":op);
				break;
			}
		}
	}
	free(name);											  /*  release copy  */
	if (result) {
		len = strlen(result);
		if (par)
			len += strlen(par);
		result = realloc(result,len+1);	 /*  adjust buffer  */
		if (result == NULL) {
			printf("unmangle_str: Out of heap (realloc)\n");
			exit(EXIT_FAILURE);
		}
		if (par) {
			strcat(result,par);
			free(par);
		}
	}
	return result;
}

/*
	Used by unmangle_str to un-mangle the type safe information.
	
	Note this returns a malloced buffer which should be freed when
	finished with if conversion takes place.
*/

char	*unmangle_parm(char *str)
{
	char	*result=NULL;
	char	*name,*s,*s1, num[4];
	static char op[20][255];
	unsigned len,i,j,no,x,flag;

	for (i = 0; i<20; i++)
		op[i][0] = 0;

	name = malloc(strlen(str)+1);					   /*  copy for safety  */
	if (name == NULL) {
		printf("unmangle_parm: Out of heap\n");
		exit(EXIT_FAILURE);
	}
	strcpy(name,str);
	s = name;
	len = j = no = 0;
	
	while (isalnum(*s)) {
		i = 0;
		while (isdigit(*s)) {
			num[i] = *s;
			i++;
			s++;
			if (i > 2) {
				printf("unmangle_parm: parsing error, number too large\n");
				exit(EXIT_FAILURE);
			}
		}
		if (i) {
			num[i] = 0;
			i = atoi(num);
			strncat(op[no],s,i);
			no++;
			s +=i;
			j++;
			continue;
		}
			
		for (i = 0; i <(sizeof(type_table)/sizeof(struct paramangle_s)); i++)
			if (type_table[i].type == *s) {
				s1 = type_table[i].desc;
				s++;
				switch (type_table[i].type) {
					case 'R': flag = i;
							  break;
							  
					case 'b':
					case 'h':
					case 'p':
					case 'P':
					case 'H': flag = i;
							  break;
							  
					case 'U':
					case 'C':
					case 'V':
							strcat(op[no],s1);
							break;
							
					case 'Z':
							--no;
							--j;
							strcat(op[no],s1);
							break;
							
					case 'A':
					 		strcat(op[no],s1);
					 		x = strlen(op[no]);
					 		while (isdigit(*s)) {
					 			op[no][x] = *s;
					 			x++;
					 			s++;
					 		}
					 		op[no][x] = 0;
					 		strcat(op[no]," ");
					 		break;
					 		
					case 'T':
							strcpy(op[no],op[(*s)-48]);
							s++;
							no++;
							j++;
							break;

					case 'D':
					case 'N':
					case 'F':
					case 'B':
							if (j) {
								*(strchr(s,'_')) = 'Z';
								strcat(op[no],"(");
								if (flag) {
									strcat(op[no],type_table[flag].desc);
									flag = 0;
								}
								strcat(op[no],s1);
								break;
							}
					default:
							strcat(op[no],s1);
							if (flag) {
								strcat(op[no],type_table[flag].desc);
								flag = 0;
							}
							no++;
							j++;
							break;
				}
				break;
			}
	}
	free(name);											  /*  release copy  */
	if (j>1) {
		for (i = 1; i < j; i++)
			len = len + strlen(op[i]) +1;
		result = malloc(len+3);					 /*  create buffer  */
		if (result == NULL) {
			printf("unmangle_parm: Out of heap\n");
			exit(EXIT_FAILURE);
		}
		strcpy(result,"(");
		for (i = 1; i < j; i++) {
			strcat(result,op[i]);
			strcat(result,",");
		}
		result[strlen(result)-1] = ')';
			
	}
	else result = NULL;
	return result;
}

/*
	unmangle_struct

	Structures need a bit of help. First unmangle all required strings and
	then pass the umangled structure member strings to this function. The
	first string is	the class name, the second string is a member name.
	The function reurns	a pointer to a buffer with the structure name in a
	more readable form if possible else NULL if the name is unchanged.

	Note this returns a malloced buffer which should be freed when
	finished with if conversion takes place.
*/

char *unmangle_struct(char *sname,char *mname)
{
	unsigned len,clen;
	char	*p=NULL,*s;

	s = malloc(strlen(sname)+3);
	if (s == NULL) {
		printf("unmangle_struct: Out of heap\n");
		exit(EXIT_FAILURE);
	}
	sprintf(s,"%s::",sname);					   /*  get full class name  */
	len = strlen(s);
	if (strncmp(s,mname,len)==0) {			 /*  matching class in member?  */
		mname += len;									  /*  get old name  */
		p = malloc(strlen(mname)+1);						   /*  copy it  */
		if (p == NULL) {
			printf("unmangle_struct: Out of heap\n");
			exit(EXIT_FAILURE);
		}
		strcpy(p,mname);
		free(mname);
	}
	else {										   /*  detect base classes  */
		if ((p=strstr(mname,"::"))!=NULL) {			  /*  find member name  */
			clen = p-mname;								  /*  class length  */
			if (strncmp(mname,p+2,clen)==0) {			 /*  if base class  */
				p = malloc(clen+1);							   /*  copy it  */
				if (p == NULL) {
					printf("unmangle_struct: Out of heap\n");
					exit(EXIT_FAILURE);
				}
				strncpy(p,mname,clen);
				p[clen] = '\0';
				free(mname);
			}
		}
	}
	free(s);
	return p;
}



